/* Chart (lattice)
 *
 * create : 95.6.23 (Fri), bgjang@csone - rewritten instead of edgePool.c
 * modify :
 * 95.6.24 (Sat), bgjang@csone - almost
 * 95.6.26 (Mon), bgjang@csone - version 0.5 alpha, Garbage Collection is
 *                              *not* tested.
 * 95.7.18 - 7.31 (Mon), bgjang
 *    Garbage Collection routine is tested and bugs are cleaned out.
 *    chartGetIth() added.
 * 95.8.7 (Mon), bgjang
 *    chart initialize part is a little bit changed to increase speed.
 *    instead of memcpy(), use for statements in chartCopy() for speed up,
 *    But the result is not so good(a little increase).
 * 95.8.8 (Tue) - 95.8.12 (Sat), bgjang
 *    Hashing data for string was deleted for speed/memory.
 *    *GetIth*() functions was changed a little for speed.
 *    In chartDelete(), path information of edge is deleted now.
 *    
 *
 *
 *--------------------------------------------------------------------
 *
 * For searching method of edges in chart, Hashing technique is used.
 * I will describe the method in detail, or say about reference in
 * each section where the Hashing is used.
 *
 *
 */
#include <stdio.h>
#include "pos.h"
#include "chart.h"

#ifndef MIN(X,Y)
#define MIN(X,Y)   (((X)<(Y))?(X):(Y))
#endif

/* data type -------------------------------------------------*/

struct hashBucket {
    int edgeIndex;
    int link;
};

struct chartPool {
    struct edge   e[MAX_EDGES_IN_CHART];
    int           used;
    long          irrApplied;
    struct hashBucket  hashHeadV[MAX_BUCKETS];    int hhN;
    struct hashBucket  hashTailV[MAX_BUCKETS];    int htN;
#if defined(LLL)
    struct hashBucket  hashEData[MAX_BUCKETS];    int heN;
#endif
};


/* global variable ------------------------------------------------*/

char chartES[1024];


/* static variable ------------------------------------------------*/

/* the last item of chart[], chart[MAX_CHART] is used for
 * garbage collection. So, it must not be used by the other
 * purpose.
 */
static struct chartPool  chart[MAX_CHART+1];
static int chartEdgeIndex[MAX_CHART];


/* static functions -------------------------------------------------*/

static void _chartGarbageCollection(/* id, pDiff */);
static int  _chartCheckID(/* id );
static int  _chartCheckEdgeID(/* id, edgeID */);
static int  _hashHeadVInsert(/* id, headV, edgeIndex */);
static int  _hashHeadVGet(/* id, headV, pEdgeIndex */);
static int  _hashHeadVGetIth(/* id, ith, headV, pEdgeIndex */);
static int  _hashHeadVMarkAsDeleted(/* id, headV, edgeID */);
static int  _hashTailVInsert(/* id, tailV, edgeIndex */);
static int  _hashTailVGet(/* id, tailV, pEdgeIndex */);
static int  _hashTailVGetIth(/* id, ith, tailV, pEdgeIndex */);
static int  _hashTailVMarkAsDeleted(/* id, tailV, edgeID */);
#if defined(LLL)
static int  _hashEDataInsert(/* id, str, edgeIndex */);
static void _hashStr2Integer(/* str, pVal */);
static int  _hashEDataGet(/* id, str, pEdgeIndex */);
static int  _hashEDataMarkAsDeleted(/* id, str, edgeID */);
#endif
#if defined(DEBUG)
static void _hashEDataDump(/* id */);
static void _hashTailVDump(/* id */);
static void _hashHeadVDump(/* id */);
#endif


/* about the exection time of the following functions
 * '0.6 real', '0.6 user', '0.0 sys' time by /usr/bin/time command
 * if (#if 0 ... #endif) section is not commented out.
 */

void chartInit()
{
    register int i, j, k;
    for( i=0; i<MAX_CHART; i++ ) {
        chart[i].used = -1;

#if 0
        /* initialize the three hash table
         */
        for( j=0; j<MAX_BUCKETS; j++ ) {
            hashHeadV[j].edgeIndex = -1; hashHeadV[j].link = -1;
            hashTailV[j].edgeIndex = -1; hashTailV[j].link = -1;
#if defined(LLL)
            hashEData[j].edgeIndex = -1; hashEData[j].link = -1;
#endif
        }

        /* initialize edges
         */
        for( j=0; j<MAX_EDGES_IN_CHART; j++ ) {
            for( k=0; k<MAX_PATH; k++ ) {
                e[j].prevEdges[k] = -1;
                e[j].nextEdges[k] = -1;
            }
        }
#endif
    }
}


int chartCreate( pId )
    int *pId;   // OUT
{
    register int i, j;
    int ind;

    /* find empty slot, and if there none, return error
     */
    for( i=0; i<MAX_CHART; i++ )
        if( chart[i].used == -1 ) break;

    if( i == MAX_CHART ) {
        sprintf(chartES, "%s : Max Chart %d is used.\n", __FILE__, i );
        return C_ERR_MAX_CHART;
    }

    ind = i;

    /* initialize one chart
     */

    /* initialize the three hash table */
    for( i=0; i<MAX_BUCKETS; i++ ) {
        chart[ind].hashHeadV[i].edgeIndex = chart[ind].hashHeadV[i].link = -1;
        chart[ind].hashTailV[i].edgeIndex = chart[ind].hashTailV[i].link = -1;
#if defined(LLL)
        chart[ind].hashEData[i].edgeIndex = chart[ind].hashEData[i].link = -1;
#endif
    }
    chart[ind].hhN = 0;
    chart[ind].htN = 0;
#if defined(LLL)
    chart[ind].heN = 0;
#endif

    /* initialize edges */
    for( i=0; i<MAX_EDGES_IN_CHART; i++ ) {
        chart[ind].e[i].headV = -1;
        chart[ind].e[i].curPrevEdge = 0;
        chart[ind].e[i].curNextEdge = 0;
/*
        for( j=0; j<MAX_PATH; j++ ) {
            chart[ind].e[i].prevEdges[j] = -1;
            chart[ind].e[i].nextEdges[j] = -1;
        }
*/
    }

    chart[ind].used = 0;

    *pId = ind;
    return C_ERR_NO;
}

int chartDestroy( id )
    int id;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    chart[id].used = -1;
    return C_ERR_NO;
}

#if defined(LOG_STATISTIC)
/* This variable is defined in MAmain.c file */
extern int maxWordLenInChart;
#endif

int chartInsertEdge( id, headV, tailV, str, epos, freq, pEdgeID )
    int id;
    int headV, tailV;
    char *str;
    enum EPOS epos;
    float freq;
    int *pEdgeID;
{
    int r;
    int edgeIndex;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    /* check current edge number and if there is free one, then
     * insert.
     */

    if( chart[id].used >= MAX_EDGES_IN_CHART ) {
        int diff;

        /* fundamemtally, in this point, garbage collection about
         * edges must be occured cause delete step do nothing but marking
         * that this edge is deleted.
         *
         * and after garabage collection, check again the free edges.
         */
        _chartGarbageCollection( id, &diff );

#if defined(DEBUG)
        printf("chartInsertEdge : GARBAGE COLLECTED, %d edges are cleared out\n", diff);
#endif

        if( chart[id].used > MAX_EDGES_IN_CHART ) {
            sprintf( chartES,"%s : Garbage was Collected but full edges are used.\n", __FILE__);
            return C_ERR_MAX_EDGES;
        }
    }


    edgeIndex = chart[id].used; chart[id].used++;

    /* fill-in edge data with handling misc. errors
     */

    if( headV >= 0 ) chart[id].e[edgeIndex].headV = headV;
    else {
        sprintf( chartES, "%s : no proper head vertex ID, %d.\n", __FILE__, headV );
        return C_ERR_HEADV;
    }
    if( tailV >= 0 ) chart[id].e[edgeIndex].tailV = tailV;
    else {
        sprintf( chartES, "%s : no proper tail vertex ID, %d.\n", __FILE__, tailV );
        return C_ERR_TAILV;
    }

#if defined(LOG_STATISTIC)
    if( str != NULL && maxWordLenInChart < strlen(str) )
        maxWordLenInChart = strlen(str);
#endif
    if( str != NULL ) strcpy( chart[id].e[edgeIndex].d.str, str );
    else strcpy( chart[id].e[edgeIndex].d.str, "" );
    chart[id].e[edgeIndex].d.epos = epos;
    chart[id].e[edgeIndex].d.freq = freq;


    /* set three hash table properly
     */
    _hashHeadVInsert( id, headV, edgeIndex );
    _hashTailVInsert( id, tailV, edgeIndex );
#if defined(LLL)
    if( str != NULL ) _hashEDataInsert( id, str, edgeIndex );
#endif

    *pEdgeID = edgeIndex;
}

/* edgeIDA -> edgeIDB
 * so chartConnect( id, edgeIDA, edgeIDB ) and
 * chartConnect( id, edgeIDB, edgeIDA ) is different!
 */
int chartConnect( id, edgeIDA, edgeIDB )
    int id;
    int edgeIDA, edgeIDB;
{
    register int i, j;
    int r;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
    if( (r = _chartCheckEdgeID(id, edgeIDA)) < 0 ) return r;
    if( (r = _chartCheckEdgeID(id, edgeIDB)) < 0 ) return r;
#endif

    /*
    for( i=0; i<MAX_PATH; i++ )
        if( chart[id].e[edgeIDA].nextEdges[i] == -1 ) break;
    */
    i = chart[id].e[edgeIDA].curNextEdge;
    if( i == MAX_PATH ) {
        sprintf( chartES, "%s : Path data for one edge is full.\n", __FILE__ );
        return C_ERR_MAX_PATH;
    }

    /*
    for( j=0; j<MAX_PATH; j++ )
        if( chart[id].e[edgeIDB].prevEdges[j] == -1 ) break;
    */
    j = chart[id].e[edgeIDB].curPrevEdge;
    if( j == MAX_PATH ) {
        sprintf( chartES, "%s : Path data for one edge is full.\n", __FILE__ );
        return C_ERR_MAX_PATH;
    }

    chart[id].e[edgeIDA].nextEdges[i] = edgeIDB;
    chart[id].e[edgeIDB].prevEdges[j] = edgeIDA;
    chart[id].e[edgeIDA].curNextEdge++;
    chart[id].e[edgeIDB].curPrevEdge++;
    return C_ERR_NO;
}

int chartPutIrrApplied( id, irrApplied )
    int id;
    long irrApplied;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif
    chart[id].irrApplied = irrApplied;
    return C_ERR_NO;
}

int chartGetIrrApplied( id, pIrrApplied )
    int id;
    long *pIrrApplied;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif
    *pIrrApplied = chart[id].irrApplied;
    return C_ERR_NO;
}

/* must be changed to using some other static variable
 */
int chartGetIth( id, ith, pEdgeID )
    int id;
    int ith;
    int *pEdgeID;    // OUT
{
    register int i;

    for( i=0; i<MAX_EDGES_IN_CHART; i++ )
        if( chart[id].e[i].headV != -1 ) {
            ith--;
            if( ith < 0 ) {
                *pEdgeID = i;
                return C_ERR_NO;
            }
        }

    return C_NO_DATA;
}

int chartGetIthByHeadV( id, ith, headV, pEdgeID )
    int id;
    int ith, headV;
    int *pEdgeID;   // OUT
{
    int r;
    int edgeID;
    /*
    int hv;
    */

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    /*
    hv = headV;
    while(1) {
        r = _hashHeadVGet( id, hv, &edgeID ); hv = -1;
    */

        r = _hashHeadVGetIth( id, ith, headV, &edgeID );
        if( r == C_NO_DATA ) return C_NO_DATA;
        *pEdgeID = edgeID;
        return C_ERR_NO;
        
    /*
        if( --ith < 0 ) {
            *pEdgeID = edgeID;
            return C_ERR_NO;
        }
    }
    */
}

int chartGetIthByTailV( id, ith, tailV, pEdgeID )
    int id;
    int ith, tailV;
    int *pEdgeID;   // OUT
{
    int r;
    int edgeID;
    /*
    int tv;
    */

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    /*
    tv = tailV;
    while(1) {
        r = _hashTailVGet( id, tv, &edgeID ); tv = -1;
    */
        r = _hashTailVGetIth( id, ith, tailV, &edgeID );
        if( r == C_NO_DATA ) return C_NO_DATA;
        *pEdgeID = edgeID;
        return C_ERR_NO;

        /*
        if( --ith < 0 ) {
            *pEdgeID = edgeID;
            return C_ERR_NO;
        }
    }
    */
}

/*
chartGetIthByHeadVTailV()
{
}
*/

#if defined(LLL)
int chartGetIthByData( id, ith, str, pEdgeID )
    int id, ith;
    char *str;
    int *pEdgeID;  // OUT
{
    int r;
    int edgeID;
    char *s;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    s = str;
    while(1) {
        r = _hashEDataGet( id, s, &edgeID ); s = NULL;
        if( r == C_NO_DATA ) return C_NO_DATA;
        if( --ith < 0 ) {
            *pEdgeID = edgeID;
            return C_ERR_NO;
        }
    }
}
#endif

/*
chartGetIthByThree()
{
}
*/

/*
chartGetByID()
{
}
*/

int chartGet( id, pEdgeID, first )
    int id;
    int *pEdgeID;
    int first;
{
    int r;
    int curEdgeID;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
#endif

    if( first > 0 )
        curEdgeID = chartEdgeIndex[id] = 0;
    else
        curEdgeID = chartEdgeIndex[id];

    /* if current edge ID marked as deleted one, then skip.
     */
    while(1) {
        if( !(curEdgeID < chart[id].used) ) return C_NO_DATA;
        if( chart[id].e[curEdgeID].headV == -1 )
            curEdgeID++;
        else break;
    }

    chartEdgeIndex[id] = curEdgeID + 1;
    *pEdgeID = curEdgeID;
    return C_ERR_NO;
}


/* mark edgeIndex's headV as -1.
 * mark three Hash Table appropriately.
 *
 * This marked edges are garbage collected when all edge was used.
 */

int chartDelete( id, edgeID )
    int id;
    int edgeID;
{
    register int i, j;
    int eId;
    int r;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return r;
    if( (r = _chartCheckEdgeID(id, edgeID)) < 0) return r;
#endif

    _hashHeadVMarkAsDeleted( id, chart[id].e[edgeID].headV, edgeID );
    _hashTailVMarkAsDeleted( id, chart[id].e[edgeID].tailV, edgeID );
#if defined(LLL)
    _hashEDataMarkAsDeleted( id, chart[id].e[edgeID].d.str, edgeID );
#endif
    chart[id].e[edgeID].headV = -1;
#if 1
    /* delete path information */
    for( i=0; i<chart[id].e[edgeID].curPrevEdge; i++ ) {
        eId = chart[id].e[edgeID].prevEdges[i];

        for( j=0; j<chart[id].e[eId].curNextEdge; j++ )
            if( chart[id].e[eId].nextEdges[j] == edgeID ) {
                chart[id].e[eId].nextEdges[j] = -1;
                break;
            }
    }
    for( i=0; i<chart[id].e[edgeID].curNextEdge; i++ ) {
        eId = chart[id].e[edgeID].nextEdges[i];

        for( j=0; j<chart[id].e[eId].curPrevEdge; j++ )
            if( chart[id].e[eId].prevEdges[j] == edgeID ) {
                chart[id].e[eId].prevEdges[j] = -1;
                break;
            }
    }
#endif
    return C_ERR_NO;
}

/* copy the contents of one chart to the other chart
 * without any changes, that is, with the deleted mark totally.
 * delete all data in chart identified by toID.
 */
int chartCopy( fromID, toID )
    int fromID, toID;
{
    int r;
#if 1
    register int i, j;
#endif

#if defined(LLL)
    if( (r = _chartCheckID(fromID)) < 0 ) return r;
    if( toID < 0 || toID > MAX_CHART ) {
        sprintf("%s : no proper chart ID, %d.\n", __FILE__, toID );
        return C_ERR_CHART_ID;
    }
#endif

    memcpy( &(chart[toID]), &(chart[fromID]), sizeof(struct chartPool) );
#if 0
    /* initialize the three hash table */
    for( i=0; i<MAX_BUCKETS; i++ ) {
        chart[toID].hashHeadV[i].edgeIndex = chart[toID].hashHeadV[i].link =
        chart[toID].hashTailV[i].edgeIndex = chart[toID].hashTailV[i].link = -1;
#if defined(LLL)
        chart[toID].hashEData[i].edgeIndex = chart[toID].hashEData[i].link = -1;
#endif
    }

    chart[toID].hhN = 0;
    chart[toID].htN = 0;
#if defined(LLL)
    chart[toID].heN = 0;
#endif

    /* initialize edges */
    for( i=0; i<MAX_EDGES_IN_CHART; i++ ) {
        chart[toID].e[i].headV = -1;
        chart[toID].e[i].curPrevEdge = 0;
        chart[toID].e[i].curNextEdge = 0;
    }

    chart[toID].used = 0;

    for( i=0, j=0; i<chart[fromID].used; i++, j++ )
        if( chart[fromID].e[i].headV != -1 ) {

            chart[toID].e[j].headV = chart[fromID].e[i].headV;
            chart[toID].e[j].tailV = chart[fromID].e[i].tailV;

            strcpy( chart[toID].e[j].d.str, chart[fromID].e[i].d.str );
            chart[toID].e[j].d.epos = chart[fromID].e[i].d.epos;
            chart[toID].e[j].d.freq = chart[fromID].e[i].d.freq;


            /* set three hash table properly
             */
            _hashHeadVInsert( toID, chart[toID].e[j].headV, j );
            _hashTailVInsert( toID, chart[toID].e[j].tailV, j );
#if defined(LLL)
            _hashEDataInsert( toID, chart[toID].e[j].d.str, j );
#endif
        } else
            chart[toID].e[j].headV = chart[fromID].e[i].headV;

    chart[toID].used = j;
#endif
    return C_ERR_NO;
}

/* garbage colleciton
 *
 * NOTICE : the connection information between edges are
 * not remained.
 */
static void _chartGarbageCollection( id, pDiff )
    int id;
    int *pDiff;  // OUT - how many deleted edge is collected.
{
    register i, j, count;
    int r;
    int first = 1;
    int edgeID, afterEdgeID;

    /* in this step, all edges must be compacted and
     * for the searching method, hash table must be reorganized
     */
    /* use MAchartCopy() alike ... 
     */

    /* clear chartFGB chart */

    for( i=0; i<MAX_BUCKETS; i++ ) {
        chart[MAX_CHART].hashHeadV[i].edgeIndex = chart[MAX_CHART].hashHeadV[i].link = -1;
        chart[MAX_CHART].hashTailV[i].edgeIndex = chart[MAX_CHART].hashTailV[i].link = -1;
#if defined(LLL)
        chart[MAX_CHART].hashEData[i].edgeIndex = chart[MAX_CHART].hashEData[i].link = -1;
#endif
    }
    chart[MAX_CHART].hhN = 0;
    chart[MAX_CHART].htN = 0;
#if defined(LLL)
    chart[MAX_CHART].heN = 0;
#endif

    for( i=0; i<MAX_EDGES_IN_CHART; i++ ) {
        chart[MAX_CHART].e[i].headV = -1;
        chart[MAX_CHART].e[i].curPrevEdge = 0;
        chart[MAX_CHART].e[i].curNextEdge = 0;
        /*
        for( j=0; j<MAX_PATH; j++ ) {
            chart[MAX_CHART].e[i].prevEdges[j] = -1;
            chart[MAX_CHART].e[i].nextEdges[j] = -1;
        }
        */
    }
    chart[MAX_CHART].used = 0;

    /* copy the contents in chart[id] to chartFGB */
    for( first=1, count=0; chartGet( id, &edgeID, first ) != C_NO_DATA; count++) {
        first = 0;
        chartInsertEdge( MAX_CHART,
                         chart[id].e[edgeID].headV,
                         chart[id].e[edgeID].tailV,
                         chart[id].e[edgeID].d.str,
                         chart[id].e[edgeID].d.epos,
                         chart[id].e[edgeID].d.freq,
                         &afterEdgeID );
    }

    *pDiff = MAX_EDGES_IN_CHART - count;

    /* recopy compacted result, chartFGB, to chart[id] */
    memcpy( &(chart[id]), &(chart[MAX_CHART]), sizeof(struct chartPool) );
}

static int _chartCheckID( id )
    int id;
{
    if( id < 0 || id > MAX_CHART ) {
        sprintf( chartES, "%s : no proper chart ID %d.\n", __FILE__, id );
        return C_ERR_CHART_ID;
    }
    if( chart[id].used == -1 ) {
        sprintf( chartES, "%s : chart must be created before using.\n", __FILE__ );
        return C_ERR_EMPTY_CHART;
    }
    return C_ERR_NO;
}

static int _chartCheckEdgeID( id, edgeID )
    int id;
    int edgeID;
{
    if( edgeID < 0 || edgeID >= chart[id].used ) {
        sprintf( chartES, "%s : no proper edge ID %d.\n", __FILE__, edgeID );
        return C_ERR_EDGE_ID;
    }
    /* if deletion step is coded, then this routine
     * must check that this edgeID is marked as deleted or not
     */
    if( chart[id].e[edgeID].headV == -1 ) return C_ERR_EDGE_ID;
    return C_ERR_NO;
}

#if defined(DEBUG)

void chartDump( id, isHH, isHT, isHE )
    int id;
    int isHH;     /* HeadV hash table dump (?) */
    int isHT;     /* TailV hash table dump (?) */
    int isHE;     /* EData hash table dump (?) */
{
    register int i, j;
    int r;

#if defined(LLL)
    if( (r = _chartCheckID(id)) < 0 ) return;
#endif

    printf("--- Dump Begin. CHART %d.\n", id);
    printf("     eID  hv   tv   pos  frq  str  prev | next\n");
    for( i=0; i<MAX_EDGES_IN_CHART; i++ ) {
        if( chart[id].e[i].headV != -1 ) {
            printf("    %3d  %3d  %3d  %4s  %f  %s  ",
              i,
              chart[id].e[i].headV,
              chart[id].e[i].tailV,
              EPOSstr[chart[id].e[i].d.epos],
              chart[id].e[i].d.freq,
              chart[id].e[i].d.str );
            for( j=0; j<chart[id].e[i].curPrevEdge; j++ )
                if( chart[id].e[i].prevEdges[j] != -1 )
                    printf("%d ", chart[id].e[i].prevEdges[j]);
            printf("| ");
            for( j=0; j<chart[id].e[i].curNextEdge; j++ )
                if( chart[id].e[i].nextEdges[j] != -1 )
                    printf("%d ", chart[id].e[i].nextEdges[j]);
            printf("\n");
        }
    }
    if( isHH > 0 ) _hashHeadVDump( id );
    if( isHT > 0 ) _hashTailVDump( id );
#if defined(LLL)
    if( isHE > 0 ) _hashEDataDump( id );
#endif
    printf("--- Dump END.\n", id);
}

#endif



/*---------------------------------------------------------------
 *
 * Hashing
 *
 *---------------------------------------------------------------
 */


/* here, two hashing method is used to searching by some attribute of
 * edge. But in the sense that (head vertex ID, tail vertex ID, string
 * for edge) is key for edge list, if we use one of the above three
 * attribute as a key to searching some edge, then the uniqueness property is
 * broken.
 * So, I use edge ID as unique key and use the above property to
 * get hash value. But the main stream of algorithm is same with
 * the referenced article.
 */

#define HASH1VERTEX(K)      ((K) % MAX_BUCKETS)

/* for vertex ID, open-addressing with linear probing is
 * used. Refer to 8.3 Hashing Techniques (P.273) section in
 * Harry R. Lewis's `Data Structures & Their Algorithms'.
 */

/* hashing for head vertex ID -> edge Index ------------------------*/

static int _hashHeadVInsert( id, headV, edgeIndex )
    int id;    // chart ID
    int headV;
    int edgeIndex;
{
    int slot, link = 1;
    int hashValue;

    hashValue = HASH1VERTEX(headV);

    slot = hashValue;

    /* if there exists and same one, then return with C_DUP */
    if( chart[id].hashHeadV[slot].edgeIndex == edgeIndex ) return C_DUP;

    /* find empty slot by linear probing
     */
    if( chart[id].hashHeadV[slot].edgeIndex != -1 ) {

        do {
            link++;
            slot = (slot + 1) % MAX_BUCKETS;
            if( chart[id].hashHeadV[slot].edgeIndex == edgeIndex ) return C_DUP;
        } while( chart[id].hashHeadV[slot].link != -1 );
    }

    /* check hash table is full or not, if full, then return error.
     * otherwise, insert data
     */
    if( chart[id].hhN == MAX_BUCKETS - 1 ) {
        sprintf("%s : Hash Table for head vertex ID is full.\n", __FILE__ );
        return C_ERR_HASH_FULL;
    }
    chart[id].hashHeadV[slot].edgeIndex = edgeIndex;
    chart[id].hashHeadV[slot].link = link;
    chart[id].hhN++;

    return C_ERR_NO;
}

static int _hashHeadVGet( id, headV, pEdgeIndex )
    int id;
    int headV;
    int *pEdgeIndex;
{
    static int link = 1;
    static int slot = 1;
    static int hv;

    if( headV != -1 ) hv = headV;

    /* if headV is negative one, then return the next edgeIndex which
     * have the same hash value as the previous one.
     */
    if( headV >= 0 ) {
        slot = HASH1VERTEX(headV);
        link = 1;
    }

    if( chart[id].hashHeadV[slot].edgeIndex == -1 ) return C_NO_DATA;

    /* find edges until empty slot is found
     */
    do {
        if( chart[id].hashHeadV[slot].link == link &&
            hv == chart[id].e[chart[id].hashHeadV[slot].edgeIndex].headV ) {
            *pEdgeIndex = chart[id].hashHeadV[slot].edgeIndex;
            slot = (slot+1) % MAX_BUCKETS;
            link++;
            return C_ERR_NO;
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashHeadV[slot].link != -1 );
    return C_NO_DATA;
}

static int _hashHeadVGetIth( id, ith, headV, pEdgeIndex )
    int id;
    int ith;
    int headV;
    int *pEdgeIndex;
{
    int link = 1;
    int slot = 1;

    slot = HASH1VERTEX(headV);
    link = 1;

    if( chart[id].hashHeadV[slot].edgeIndex == -1 ) return C_NO_DATA;

    /* find edges until empty slot is found
     */
    do {
        if( chart[id].hashHeadV[slot].link == link &&
            headV == chart[id].e[chart[id].hashHeadV[slot].edgeIndex].headV ) {
            ith--;
            if( ith < 0 ) {
                *pEdgeIndex = chart[id].hashHeadV[slot].edgeIndex;
                slot = (slot+1) % MAX_BUCKETS;
                link++;
                return C_ERR_NO;
            }
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashHeadV[slot].link != -1 );

    return C_NO_DATA;
}

/* deleted edge's link value is set to 0.
 * initially, edge's link value is -1. and the other
 * normal case (edge data is existed), then edge's link value
 * is positive.
 *
 * so we must see empty slot by checking link value is -1 or not.
 * and valid one or not, after see link value is positive or not.
 */
static int _hashHeadVMarkAsDeleted( id, headV, edgeID )
    int id;
    int headV;
    int edgeID;
{
    int slot, link;

    slot = HASH1VERTEX(headV);
    link = 1;

    do {
        if( chart[id].hashHeadV[slot].link == link ) {
            if( edgeID == chart[id].hashHeadV[slot].edgeIndex ) {
                chart[id].hashHeadV[slot].edgeIndex = -1;
                chart[id].hashHeadV[slot].link = 0;
                return C_ERR_NO;
            }
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashHeadV[slot].link != -1 );
    return C_NO_DATA;
}

#if defined(DEBUG)

static void _hashHeadVDump( id )
    int id;
{
    register int i;
    int K, eInd, headV;

    printf("--- Dump Begin. Hash Table for head vertex ID\n");
    printf("    slot   h1  link  edgeID\n");
    for( i=0; i<MAX_BUCKETS; i++ ) {
        if( chart[id].hashHeadV[i].edgeIndex >= 0 ) {

            eInd = chart[id].hashHeadV[i].edgeIndex;
            headV = chart[id].e[eInd].headV;

            printf("    %4d  %3d  %4d  %6d\n",
                i, HASH1VERTEX(headV),
                chart[id].hashHeadV[i].link,
                chart[id].hashHeadV[i].edgeIndex );
        }
    }
    printf("--- Dump End.\n");
}
#endif


/* hashing for tail vertex ID -> edge Index ------------------------*/

/* all is same with that for head vertex ID.
 */
static int _hashTailVInsert( id, tailV, edgeIndex )
    int id;    // chart ID
    int tailV;
    int edgeIndex;
{
    int slot, link = 1;
    int hashValue;

    hashValue = HASH1VERTEX(tailV);

    slot = hashValue;

    if( chart[id].hashTailV[slot].edgeIndex == edgeIndex ) return C_DUP;

    if( chart[id].hashTailV[slot].edgeIndex != -1 ) {
        do {
            link++;
            slot = (slot + 1) % MAX_BUCKETS;
            if( chart[id].hashTailV[slot].edgeIndex == edgeIndex ) return C_DUP;
        } while( chart[id].hashTailV[slot].link != -1 );
    }

    if( chart[id].htN == MAX_BUCKETS - 1 ) {
        sprintf("%s : Hash Table for tail vertex ID is full.\n", __FILE__ );
        return C_ERR_HASH_FULL;
    }
    chart[id].hashTailV[slot].edgeIndex = edgeIndex;
    chart[id].hashTailV[slot].link = link;
    chart[id].htN++;

    return C_ERR_NO;
}

static int _hashTailVGet( id, tailV, pEdgeIndex )
    int id;
    int tailV;
    int *pEdgeIndex;
{
    static int link = 1;
    static int slot = 1;
    static int tv;

    if( tailV != -1 ) tv = tailV;

    if( tailV >= 0 ) {
        slot = HASH1VERTEX(tailV);
        link = 1;
    }

    if( chart[id].hashTailV[slot].edgeIndex == -1 ) return C_NO_DATA;

    do {
        if( chart[id].hashTailV[slot].link == link /* &&
            tv == chart[id].e[chart[id].hashTailV[slot].edgeIndex].tailV */) {
            *pEdgeIndex = chart[id].hashTailV[slot].edgeIndex;
            slot = (slot+1) % MAX_BUCKETS;
            link++;
            return C_ERR_NO;
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashTailV[slot].link != -1 );
    return C_NO_DATA;
}

static int _hashTailVGetIth( id, ith, tailV, pEdgeIndex )
    int id;
    int ith;
    int tailV;
    int *pEdgeIndex;
{
    int link = 1;
    int slot = 1;

    slot = HASH1VERTEX(tailV);
    link = 1;

    if( chart[id].hashTailV[slot].edgeIndex == -1 ) return C_NO_DATA;

    do {
        if( chart[id].hashTailV[slot].link == link /* &&
            tv == chart[id].e[chart[id].hashTailV[slot].edgeIndex].tailV */) {
            ith--;
            if( ith < 0 ) {
                *pEdgeIndex = chart[id].hashTailV[slot].edgeIndex;
                slot = (slot+1) % MAX_BUCKETS;
                link++;
                return C_ERR_NO;
            }
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashTailV[slot].link != -1 );

    return C_NO_DATA;
}

static int _hashTailVMarkAsDeleted( id, tailV, edgeID )
    int id;
    int tailV;
    int edgeID;
{
    int slot, link;

    slot = HASH1VERTEX(tailV);
    link = 1;

    do {
        if( chart[id].hashTailV[slot].link == link ) {
            if( edgeID == chart[id].hashTailV[slot].edgeIndex ) {
                chart[id].hashTailV[slot].edgeIndex = -1;
                chart[id].hashTailV[slot].link = 0;
                return C_ERR_NO;
            }
        }
        slot = (slot+1) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashTailV[slot].link != -1 );
    return C_NO_DATA;
}

#if defined(DEBUG)

static void _hashTailVDump( id )
    int id;
{
    register int i;
    int K, eInd, tailV;

    printf("--- Dump Begin. Hash Table for tail vertex ID\n");
    printf("    slot   h1  link  edgeID\n");
    for( i=0; i<MAX_BUCKETS; i++ ) {
        if( chart[id].hashTailV[i].edgeIndex >= 0 ) {

            eInd = chart[id].hashTailV[i].edgeIndex;
            tailV = chart[id].e[eInd].tailV;

            printf("    %4d  %3d  %4d  %6d\n",
                i, HASH1VERTEX(tailV),
                chart[id].hashTailV[i].link,
                chart[id].hashTailV[i].edgeIndex );
        }
    }
    printf("--- Dump End.\n");
}
#endif


/* hashing for string in edge data -> edge Index -------------------*/

#if defined(LLL)

#define HASH1EDATA(K)   ((K) % MAX_BUCKETS)
#define HASH2EDATA(K)   (1 + ((K) % (MAX_BUCKETS-2)))

/* for string data in edge data, open-addressing with double hash
 * is used. Refer to Knuth (The Art of Programming, Vol.3) (6.4) Algorithm.
 */
static int _hashEDataInsert( id, str, edgeIndex )
    int id;
    char *str;
    int edgeIndex;
{
    int slot, link = 1;
    int K, hashValue1, hashValue2;

    _hashStr2Integer( str, &K );

    /* first hash */
    hashValue1 = HASH1EDATA(K);

    slot = hashValue1;

    /* if there exists and the same one, then return with C_DUP */
    if( chart[id].hashEData[slot].edgeIndex == edgeIndex ) return C_DUP;

    /* if not empty, then find empty slot using second hash
     */
    if( chart[id].hashEData[slot].edgeIndex != -1 ) {
        /* second hash */
        hashValue2 = HASH2EDATA(K);

        do {
            link++;
            slot = (slot + hashValue2) % MAX_BUCKETS;
            if( chart[id].hashEData[slot].edgeIndex == edgeIndex ) return C_DUP;
        } while( chart[id].hashEData[slot].link != -1 );
    }

    /* check hash table is full or not, if full, then return error.
     * otherwise, insert data
     */
    if( chart[id].heN == MAX_BUCKETS - 1 ) {
        sprintf("%s : Hash Table for string in edge data is full.\n", __FILE__ );
        return C_ERR_HASH_FULL;
    }
    chart[id].hashEData[slot].edgeIndex = edgeIndex;
    chart[id].hashEData[slot].link = link;
    chart[id].heN++;

    return C_ERR_NO;
}

/* string to integer value. Algorithm is simple, we look the first
 * several character and sum those. If str is too short, then
 * see all.
 */
static void _hashStr2Integer( str, pVal )
    char *str;
    int *pVal;
{
#define MAX_LOOK 7
    register int i;
    int N;
    int value = 0;

    N = MIN( strlen(str), MAX_LOOK );

    for( i=0; i<N; i++ )
        value += (unsigned char) str[i];

    *pVal = value;
}

static int _hashEDataGet( id, str, pEdgeIndex )
    int id;
    char *str;
    int *pEdgeIndex;
{
    /* if str is NULL, then return the next edgeIndex which have the
     * same hash value as the previous one.
     */
    static int link = 1;
    static int slot = 1;
    static int K = 1;
    static char s[128];
    int hashValue2;

    if( str != NULL ) strcpy( s, str );

    /* if headV is negative one, then return the next edgeIndex which
     * have the same hash value as the previous one.
     */
    if( str != NULL ) {
        _hashStr2Integer( str, &K );
        slot = HASH2EDATA(K);
        link = 1;
    }

    if( chart[id].hashEData[slot].edgeIndex == -1 ) return C_NO_DATA;

    hashValue2 = HASH2EDATA(K);

    /* find edges until empty slot is found
     */
    do {
        if( chart[id].hashEData[slot].link == link &&
            !strcmp( s, chart[id].e[chart[id].hashEData[slot].edgeIndex].d.str) ) {
            *pEdgeIndex = chart[id].hashEData[slot].edgeIndex;
            slot = (slot+hashValue2) % MAX_BUCKETS;
            link++;
            return C_ERR_NO;
        }
        slot = (slot+hashValue2) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashEData[slot].link != -1 );
    return C_NO_DATA;
}

static int _hashEDataMarkAsDeleted( id, str, edgeID )
    int id;
    char *str;
    int edgeID;
{
    int slot, link;
    int K, hashValue2;

    _hashStr2Integer( str, &K );

    slot = HASH1EDATA(K);
    hashValue2 = HASH2EDATA(K);

    do {
        if( chart[id].hashEData[slot].link == link ) {
            if( edgeID == chart[id].hashEData[slot].edgeIndex ) {
                chart[id].hashEData[slot].edgeIndex = -1;
                chart[id].hashEData[slot].link = 0;
                return C_ERR_NO;
            }
        }
        slot = (slot+hashValue2) % MAX_BUCKETS;
        link++;
    } while( chart[id].hashEData[slot].link != -1 );
    return C_NO_DATA;
}

#if defined(DEBUG)

static void _hashEDataDump( id )
    int id;
{
    register int i;
    int K, eInd;

    printf("--- Dump Begin. Hash Table for string in edge data\n");
    printf("    slot   h1   h2  link  edgeID\n");
    for( i=0; i<MAX_BUCKETS; i++ ) {
        if( chart[id].hashEData[i].edgeIndex >= 0 ) {

            eInd = chart[id].hashEData[i].edgeIndex;
            _hashStr2Integer( chart[id].e[eInd].d.str, &K );

            printf("    %4d  %3d  %3d  %4d  %6d\n",
                i, HASH1EDATA(K), HASH2EDATA(K),
                chart[id].hashEData[i].link,
                chart[id].hashEData[i].edgeIndex );
        }
    }
    printf("--- Dump End.\n");
}
#endif


#endif /* of #if defined(LLL) */


/* edge functions ---------------------------------------------*/

/* routines to retrieve the value from edge
 */

int edgeStrOf( id, edgeID, str )
    int id, edgeID;
    char *str;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    strcpy( str, chart[id].e[edgeID].d.str );
    return C_ERR_NO;
}

int edgeEPOSOf( id, edgeID, pEpos )
    int id, edgeID;
    enum EPOS *pEpos;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pEpos = chart[id].e[edgeID].d.epos;
    return C_ERR_NO;
}

int edgePOSOf( id, edgeID, pPos )
    int id, edgeID;
    enum POS *pPos;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pPos = (enum POS) chart[id].e[edgeID].d.epos;
    return C_ERR_NO;
}

int edgeFreqOf( id, edgeID, pFreq )
    int id, edgeID;
    float *pFreq;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pFreq = chart[id].e[edgeID].d.freq;
    return C_ERR_NO;
}

/* for T-type chart, IE value is stored in epos variable.
 * look MAchartInsertT() functions.
 */
int edgeIorEOf( id, edgeID, pIorE )
    int id, edgeID;
    enum IE *pIorE;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pIorE = (enum IE) chart[id].e[edgeID].d.epos;
    return C_ERR_NO;
}

int edgeLmRmOf( id, edgeID, pLm, pRm )
    int id, edgeID;
    char *pLm, *pRm;
{
    int r;
    char lm, rm;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    sscanf( chart[id].e[edgeID].d.str, "%c %c", &lm, &rm );
    *pLm = lm; *pRm = rm;
    return C_ERR_NO;
}

int edgeHeadVOf( id, edgeID, pHeadV )
    int id, edgeID;
    int *pHeadV;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pHeadV = chart[id].e[edgeID].headV;
    return C_ERR_NO;
}

int edgeTailVOf( id, edgeID, pTailV )
    int id, edgeID;
    int *pTailV;
{
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    /* error is checked in insert step, chartInsert() */
    *pTailV = chart[id].e[edgeID].tailV;
    return C_ERR_NO;
}

int edgeIthNextOf( id, edgeID, ith, pNext )
    int id, edgeID;
    int ith, *pNext;
{
    register int i;
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    if( chart[id].e[edgeID].curNextEdge <= ith ) return C_NO_DATA;
    else {
        while( chart[id].e[edgeID].nextEdges[ith] == -1 ) {
            if( i > chart[id].e[edgeID].curNextEdge-1 ) return C_NO_DATA;
            for( i=ith; i<chart[id].e[edgeID].curNextEdge-1; i++ )
                chart[id].e[edgeID].nextEdges[i] =
                               chart[id].e[edgeID].nextEdges[i+1];
            chart[id].e[edgeID].curNextEdge--;
        }
        *pNext = chart[id].e[edgeID].nextEdges[ith];
        return C_ERR_NO;
    }
}

int edgeIthPrevOf( id, edgeID, ith, pPrev )
    int id, edgeID;
    int ith, *pPrev;
{
    register int i;
    int r;
#if defined(LLL)
    if( (r = _chartCheckID( id )) < 0 ) return r;
    if( (r = _chartCheckEdgeID( id, edgeID )) < 0 ) return r;
#endif

    if( chart[id].e[edgeID].curPrevEdge <= ith ) return C_NO_DATA;
    else {
        while( chart[id].e[edgeID].prevEdges[ith] == -1 ) {
            if( i > chart[id].e[edgeID].curPrevEdge-1) return C_NO_DATA;
            for( i=ith; i<chart[id].e[edgeID].curPrevEdge-1; i++ )
                chart[id].e[edgeID].prevEdges[i] =
                               chart[id].e[edgeID].prevEdges[i+1];
            chart[id].e[edgeID].curPrevEdge--;
        }
        *pPrev = chart[id].e[edgeID].prevEdges[ith];
        return C_ERR_NO;
    }
}
